import { deserialize } from './impl'
import { Deserialize, IDeserializer, IMapAccess, IterableAccess, IVisitor, MapAccess } from './interface'
import { GlobalRegistry, Registry } from '../registry'
import { IterResult } from '../utils'

export function forward(value: any, into: any, registry: Registry = GlobalRegistry) {
  const forwarder = new Forwarder(value)
  return deserialize(forwarder, into, registry)
}

type Entry = [PropertyKey, any]

class ForwardMapAccess extends MapAccess {
  private readonly _entries: Entry[]
  private index: number = -1

  constructor(entries: Entry[]) {
    super()
    this._entries = entries
  }

  static fromObject(value: Record<PropertyKey, any>) {
    return new this(Object.entries(value))
  }

  nextKey<T, K extends Deserialize<T>>(seed?: K): IteratorResult<T> {
    this.index += 1

    if (this.index >= this._entries.length) {
      return IterResult.Done()
    } else {
      const key = this._entries[this.index][0]
      const value = seed != null ? seed(new Forwarder(key)) : key as T
      return IterResult.Next(value)
    }
  }

  nextValue<T, V extends Deserialize<T>>(seed?: V): IteratorResult<T> {
    if (this.index >= this._entries.length) {
      return IterResult.Done()
    } else {
      const value = this._entries[this.index][1]
      const deser = seed != null ? seed(value) : value
      return IterResult.Next(deser)
    }
  }
}

class ForwardIterableAccess<T> extends IterableAccess {
  private readonly elements: T[]
  private index: number = -1

  constructor(elements: T[]) {
    super()
    this.elements = elements
  }

  nextElement<T, D extends Deserialize<T>>(seed?: D): IteratorResult<T> {
    this.index += 1

    if (this.index >= this.elements.length) {
      return IterResult.Done()
    } else {
      const element = this.elements[this.index]
      const deser = seed != null ? seed(new Forwarder(element)) : element
      return IterResult.Next(deser as T)
    }
  }
}

export class Forwarder implements IDeserializer {
  private readonly value: any

  constructor(value: any) {
    this.value = value
  }

  deserializeAny<T>(visitor: IVisitor<T>): T {
    switch (typeof this.value) {
      case 'string': return this.deserializeString(visitor)
      case 'number': return this.deserializeNumber(visitor)
      case 'bigint': return this.deserializeBigInt(visitor)
      case 'boolean': return this.deserializeBoolean(visitor)
      case 'symbol': return this.deserializeSymbol(visitor)
      case 'undefined': return this.deserializeNull(visitor)
      case 'object': {
        switch (true) {
          case Array.isArray(this.value): return this.deserializeIterable(visitor)
          default: return this.deserializeObject(visitor)
        }
      }
      case 'function': return this.deserializeFunction(visitor)
    }
  }

  deserializeBoolean<T>(visitor: IVisitor<T>): T {
    return visitor.visitBoolean(this.value)
  }

  deserializeNumber<T>(visitor: IVisitor<T>): T {
    return visitor.visitNumber(this.value)
  }

  deserializeBigInt<T>(visitor: IVisitor<T>): T {
    return visitor.visitBigInt(this.value)
  }

  deserializeString<T>(visitor: IVisitor<T>): T {
    return visitor.visitString(this.value)
  }

  deserializeSymbol<T>(visitor: IVisitor<T>): T {
    return visitor.visitSymbol(this.value)
  }

  deserializeNull<T>(visitor: IVisitor<T>): T {
    return visitor.visitNull()
  }

  deserializeObject<T>(visitor: IVisitor<T>): T {
    return visitor.visitObject(ForwardMapAccess.fromObject(this.value) as IMapAccess)
  }

  deserializeIterable<T>(visitor: IVisitor<T>): T {
    return visitor.visitIterable(new ForwardIterableAccess(this.value))
  }

  deserializeFunction<T>(_visitor: IVisitor<T>): T {
    throw new Error('Method not implemented.')
  }
}

