import { IterResult, Nullable } from '../utils'

export interface IMapAccess {
  nextKey<T>(seed?: Deserialize<T>): IteratorResult<T>
  nextValue<T>(seed?: Deserialize<T>): IteratorResult<T>
  nextEntry<K, V>(kseed?: Deserialize<K>, vseed?: Deserialize<V>): IteratorResult<[K, V]>
  sizeHint?(): Nullable<number>
  entries<T>(seed?: Deserialize<T>): Iterator<T>
  values<T>(seed?: Deserialize<T>): Iterator<T>
  entries<K, V>(kseed: Deserialize<K>, vseed: Deserialize<V>): Iterator<[K, V]>
  [Symbol.iterator]<K, V>(): Iterator<[K, V]>
}

export abstract class MapAccess {
  abstract nextKey<T>(seed?: Deserialize<T>): IteratorResult<T>
  abstract nextValue<T>(seed?: Deserialize<T>): IteratorResult<T>

  nextEntry<K, V>(kseed?: Deserialize<K>, vseed?: Deserialize<V>): IteratorResult<[K, V]> {
    const key = this.nextKey(kseed) as IteratorResult<K>

    if (!key.done) {
      const value = this.nextValue(vseed) as IteratorResult<V>

      if (!value.done) {
        return IterResult.Next([key.value, value.value])
      }
    }

    return IterResult.Done()
  }

  private *generate<T>(next: () => IteratorResult<T>): Iterator<T> {
    let item

    while ((item = next()) && !item.done) {
      yield item.value
    }
  }

  keys<T>(seed?: Deserialize<T>): Iterator<T> {
    return this.generate<T>(this.nextKey.bind(this, seed) as any)
  }

  values<T>(seed?: Deserialize<T>): Iterator<T> {
    return this.generate(this.nextValue.bind(this, seed) as any)
  }

  entries<K, V>(kseed?: Deserialize<K>, vseed?: Deserialize<V>): Iterator<[K, V]> {
    return this.generate(this.nextEntry.bind(this, kseed, vseed) as any)
  }

  [Symbol.iterator]<K, V>(): Iterator<K, V> {
    return this.entries() as Iterator<K, V>
  }
}

export interface IIterableAccess {
  nextElement<T>(seed: Deserialize<T>): IteratorResult<T>
  sizeHint?(): Nullable<number>
  [Symbol.iterator]<T>(): Iterator<T>
}

export abstract class IterableAccess implements IIterableAccess {
  abstract nextElement<T>(seed: Deserialize<T>): IteratorResult<T>

  [Symbol.iterator]<T>(): Iterator<T> {
    return {
      next: this.nextElement.bind(this)
    }
  }
}

const VisitorMethods = Object.freeze([
  'visitBoolean',
  'visitNumber',
  'visitBigInt',
  'visitString',
  'visitSymbol',
  'visitNull',
  'visitObject',
  'visitIterable'
] as const)

export function isVisitor<T>(visitor: any): visitor is IVisitor<T> {
  return VisitorMethods.every(method => method in visitor)
}

export interface IVisitor<T> {
  visitBoolean(value: boolean): T
  visitNumber(value: number): T
  visitBigInt(value: bigint): T
  visitString(value: string): T
  visitSymbol(value: symbol): T
  visitNull(): T
  visitObject(access: IMapAccess): T
  visitIterable(access: IIterableAccess): T
}

export interface IDeserializer {
  deserializeAny<T>(visitor: Partial<IVisitor<T>>): T
  deserializeBoolean<T>(visitor: Partial<IVisitor<T>>): T
  deserializeNumber<T>(visitor: Partial<IVisitor<T>>): T
  deserializeBigInt<T>(visitor: Partial<IVisitor<T>>): T
  deserializeString<T>(visitor: Partial<IVisitor<T>>): T
  deserializeSymbol<T>(visitor: Partial<IVisitor<T>>): T
  deserializeNull<T>(visitor: Partial<IVisitor<T>>): T
  deserializeObject<T>(visitor: Partial<IVisitor<T>>): T
  deserializeIterable<T>(visitor: Partial<IVisitor<T>>): T
  deserializeFunction<T>(visitor: Partial<IVisitor<T>>): T
}

export interface Deserialize<T> {
  (deserializer: IDeserializer): T
}

